﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

namespace Eliza
{
    /// <summary>
    /// A rule that associates decomposition of an input string with a collection of
    /// reassembled responses.
    /// </summary>
    public class DecompReassemblyRule
    {
        #region Constructor

        /// <summary>
        /// Initializes a new instance of the <see cref="DecompReassemblyRule"/> class.
        /// </summary>
        /// <param name="decomp">How the input string should be decomposed.</param>
        /// <param name="reassembledResponses">The reassembled responses.</param>
        public DecompReassemblyRule
            (string decomp,
             List<string> reassembledResponses)
        {
            this._decomp = decomp;
            this._reassembledResponses = new List<string>();
            foreach(string reassembledResponse in reassembledResponses) {
                this._reassembledResponses.Add(reassembledResponse);
            }
            this._nextResponseIndex = 0;
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Determines whether this instance can decompose the specified input.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <param name="synonyms">The synonyms.</param>
        /// <returns>
        /// <see langword="true"/> if this instance can decompose the specified input; otherwise,
        /// <see langword="false"/>.
        /// </returns>
        public bool CanDecompose
            (string input,
             Dictionary<string, List<string>> synonyms)
        {
            //Debug.WriteLine(string.Format("  DecompReassamblyRule.CanDecompose: Checking rule \"{0}\"...", this._decomp));
            //Console.WriteLine(string.Format("  DecompReassamblyRule.CanDecompose: Checking rule \"{0}\"...", this._decomp));

            // STEP 1:
            //
            // Extract each element in the decomposition definition.
            // If the decomposition was "* BLAH BLAH * BLAH BLAH BLAH *", the extracted elements would be:
            //    element 1 = *
            //    element 2 = BLAH BLAH
            //    element 3 = *
            //    element 4 = BLAH BLAH BLAH
            //    element 5 = *

            // For each word in the decomposition definition...
            List<string> decompElements = new List<string>();
            foreach(string word in this._decomp.Split(null)) {

                // If (1) this is the first word -or-
                //    (2) the word contains "*" -or-
                //    (3) the previously extracted element is "*"
                if (!decompElements.Any() ||
                     word.Contains("*") ||
                     decompElements.LastOrDefault() == "*") {

                    // Add the word to the list of elements
                    decompElements.Add(word);
                }
                else {
                    // Otherwise, concatenate the word to the last element in the list
                    string lastDecompElement = decompElements.Last();
                    lastDecompElement += " " + word;
                    decompElements[decompElements.Count - 1] = lastDecompElement;
                }
            }

            // Pad each extracted key phrase element with a space
            for(int i=0; (i < decompElements.Count); i++) {
                if (decompElements[i] != "*") {
                    decompElements[i] = string.Format(" {0} ", decompElements[i]);
                }
            }

            // STEP 2:
            //
            // Next, for each decomposition element, find the corresponding text fragment
            // in the user's input.  Store these fragments in the _textElements list.

            // Pad the input with spaces
            input = string.Format(" {0} ", input.Trim());

            // Do some initialization
            int start = 0;                              // the point where we are in the user's input
            int decompEleIndex = 0;                     // which decomposition element we're considering
            this._textElements = new List<string>();    // create a fresh _textElements list

            // For each decomposition element...
            while (decompEleIndex < decompElements.Count) {
                string decompEle = decompElements[decompEleIndex];

                // If the element is "*" (which represents optional text), gather this text
                // and store it in the _textElements list
                if (decompEle == "*") {
                    if (decompEleIndex == decompElements.Count - 1) {
                        string textAfterLastMatchedDecompEle = input.Substring(start);
                        this._textElements.Add(textAfterLastMatchedDecompEle);
                    }
                }
                else {

                    // Otherwise, the element is a key phrase/synonym we're trying to match.
                    bool isSynonym = decompEle.Trim().Contains("@");
                    int decompEleStart = -1;
                    string matchedSynonym = null;

                    if (isSynonym) {

                        // Get the snyonym (i.e. the word starting with @) in the key phrase
                        string[] wordsInKeyPhrase = decompEle.Split();
                        string theSynonym = null;
                        foreach(string word in wordsInKeyPhrase) {
                            if (word.StartsWith("@")) {
                                theSynonym = word;
                                break;
                            }
                        }

                        // Check if any of the element's synonyms are in the user's input
                        foreach(string synonym in synonyms[theSynonym]) {
                            decompEleStart = input.IndexOf(synonym, start);
                            if (decompEleStart != -1) {
                                matchedSynonym = synonym;
                                break;
                            }
                        }
                    }
                    else {
                        // Check if the element is in the user's input
                        decompEleStart = input.IndexOf(decompEle, start);
                    }

                    // If this element isn't in the user's input, return false as we don't
                    // need to continue trying to match subsequent elements.
                    if (decompEleStart == -1) {
                        return false;
                    }

                    // Since the key phrase/synonym matched, store the text before the matched element
                    // as well as the matched element itself, in the _textElements list
                    string textBeforeMatchedDecompEle = input.Substring(start, decompEleStart - start);
                    this._textElements.Add(textBeforeMatchedDecompEle);
                    this._textElements.Add(isSynonym ? matchedSynonym : decompEle);

                    // Reposition from where we want to continue searching for elements
                    start = decompEleStart + (isSynonym ? matchedSynonym.Length : decompEle.Length - 1);
                }
                decompEleIndex++;
            }

            // Since we haven't returned "false" from the previous loop, we've found
            // all elements, so return true
            return true;
        }

        /// <summary>
        /// Generates a response to the user's input.
        /// </summary>
        /// <param name="input">The user's input.</param>
        /// <returns>
        /// The response.
        /// </returns>
        public string GenerateResponse
            (string input)
        {
            // Select the next reponse
            string response = this._reassembledResponses[this._nextResponseIndex];
            this._nextResponseIndex++;
            if (this._nextResponseIndex == this._reassembledResponses.Count) {
                this._nextResponseIndex = 0;
            }

            // Replace any numbered placeholders with the corresponding text elements, after
            // transforming them
            for (int i=0; (i < this._textElements.Count); i++) {
                string placeHolder = string.Format("({0})", i + 1);
                if (response.Contains(placeHolder)) {
                    string transformedInputElement = StringTransformer.PostTransform(this._textElements[i]);
                    response = response.Replace(placeHolder, transformedInputElement);
                }
            }

            // Reduce double spaces with a single space and remove spaces before punctuation
            while(response.Contains("  ")) {
                response = response.Replace("  ", " ");
            }
            response = response.Replace(" ?", "?");
            response = response.Replace(" !", "!");
            response = response.Replace(" .", ".");
            response = response.Replace(" ,", ",");
            response = response.Trim();

            // Add an extra space after embedded punctuation
            response = response.Replace("? ", "?  ");
            response = response.Replace("! ", "!  ");
            response = response.Replace(". ", ".  ");

            // Return the response
            return response;
        }

        #endregion

        #region Fields

        /// <summary>
        /// How the input string should be decomposed.
        /// </summary>
        private string _decomp;

        /// <summary>
        /// The text elements in the user's input.
        /// </summary>
        private List<string> _textElements;

        /// <summary>
        /// The collection of reassembled responses.
        /// </summary>
        private List<string> _reassembledResponses;

        /// <summary>
        /// Index of the next reassembled response to return.
        /// </summary>
        private int _nextResponseIndex;

        #endregion
    }
}
